from abc import ABC, abstractmethod
from typing import Dict

from django.core.mail import EmailMultiAlternatives
from django.template import loader

from .enums import NotificationType
from .settings import otp_settings


class BaseNotifier(ABC):
    notification_type: NotificationType

    @abstractmethod
    def notify(self, email: str, token: str, extra_context: Dict, **kwargs) -> None:  # nocover
        raise NotImplementedError()


class EmailNotifier(BaseNotifier):
    notification_type = NotificationType.EMAIL

    def notify(self, email: str, token: str, extra_context: Dict, **kwargs) -> None:
        email_config = otp_settings.EMAIL_CONFIG
        context = {"token": token, **extra_context}

        request = context.pop("request", None)
        if request:
            context["base_url"] = request.build_absolute_uri("/")

        email = EmailMultiAlternatives(
            subject=email_config["SUBJECT"],
            body=loader.render_to_string(email_config["TEMPLATE_PATH"], context),
            to=[email],
        )
        email.content_subtype = 'html'
        email.send()



from typing import Any, Dict, Optional

from django.conf import settings
from django.test.signals import setting_changed
from django.utils.translation import gettext_lazy as _

from .enums import NotificationType, TokenType

DEFAULTS = {
    "ENABLE_OTP_AUTHENTICATION": True,
    "TOKEN_LENGTH": 6,
    "TOKEN_SECRET": "ABCDEFGHIJKLMNOPQRSTUVWXYZ234567",
    "TOTP_TOKEN_TTL": 300,
    "DEFAULT_TOKEN_TYPE": TokenType.TOTP,
    "DEFAULT_NOTIFICATION_TYPE": NotificationType.EMAIL,
    "EMAIL_CONFIG": {
        "SUBJECT": _("Verification Token"),
        "TEMPLATE_PATH": "django_drf_otp/email.html",
    },
}


class Settings:
    def __init__(self, defaults: Dict, user_settings: Optional[Dict]) -> None:
        self._user_settings = user_settings
        self.defaults = defaults or DEFAULTS

    @property
    def user_settings(self):
        if not hasattr(self, "_user_settings") or self._user_settings is None:
            self._user_settings = getattr(settings, "OTP_SETTINGS", {})
        return self._user_settings

    def __getattr__(self, attr: str) -> Any:
        if attr not in self.defaults:
            raise AttributeError("Invalid API setting: '%s'" % attr)

        try:
            val = self.user_settings[attr]
        except KeyError:
            val = self.defaults[attr]
        return val

    def reload(self):
        if hasattr(self, '_user_settings'):
            delattr(self, '_user_settings')


otp_settings = Settings(DEFAULTS, None)


def reload_api_settings(*args, **kwargs):
    setting = kwargs['setting']
    if setting == 'OTP_SETTINGS':
        otp_settings.reload()


setting_changed.connect(reload_api_settings)



import base64
import typing
from abc import ABC, abstractmethod
from typing import Optional

from django.contrib.auth.models import AbstractUser
from pyotp import TOTP

from .enums import TokenType
from .models import VerificationToken
from .settings import otp_settings


class BaseOTPStrategy(ABC):
    token_type: TokenType

    @abstractmethod
    def generate_token(self, extra_secret_chars: Optional[str] = None) -> str:  # nocover
        raise NotImplementedError()

    @abstractmethod
    def verify_token(self, token: str, extra_secret_chars: Optional[str] = None) -> bool:  # nocover
        raise NotImplementedError()

    @abstractmethod
    def generate_token_for_user(self, user: typing.Type[AbstractUser]) -> VerificationToken:  # nocover
        raise NotImplementedError()

    @staticmethod
    def create_verification_token(
        user: AbstractUser, token: str, token_type: TokenType, secret_chars: str
    ) -> VerificationToken:
        return VerificationToken.objects.create(
            user=user, token=token, token_type=token_type, secret_chars=secret_chars
        )

    @staticmethod
    def deactivate_old_verification_tokens(verification_token: VerificationToken) -> None:
        VerificationToken.objects.filter(user=verification_token.user, is_active=True).exclude(
            id=verification_token.id
        ).update(is_active=False)


class TOTPStrategy(BaseOTPStrategy):
    token_type = TokenType.TOTP

    def generate_token(self, extra_secret_chars: Optional[str] = None) -> str:
        token_class = self._get_token_class(extra_secret_chars=extra_secret_chars)
        return token_class.now()

    def verify_token(self, token: str, extra_secret_chars: Optional[str] = None) -> bool:
        token_class = self._get_token_class(extra_secret_chars=extra_secret_chars)
        return token_class.verify(otp=token)

    def generate_token_for_user(self, user: AbstractUser) -> VerificationToken:
        secret_chars = VerificationToken.generate_secret_chars(user=user)

        token = self.generate_token(extra_secret_chars=secret_chars)
        verification_token = self.create_verification_token(
            user=user,
            token=token,
            token_type=self.token_type,
            secret_chars=secret_chars,
        )
        self.deactivate_old_verification_tokens(verification_token=verification_token)
        return verification_token

    def _get_token_class(self, extra_secret_chars: Optional[str] = None) -> TOTP:
        base32_secret_chars = self._get_base32_secret(extra_secret_chars=extra_secret_chars)
        return TOTP(s=base32_secret_chars, digits=otp_settings.TOKEN_LENGTH, interval=otp_settings.TOTP_TOKEN_TTL)

    def _get_base32_secret(self, extra_secret_chars: Optional[str] = None) -> str:
        base32_secret_chars = otp_settings.TOKEN_SECRET
        if extra_secret_chars:
            base32_secret_chars += extra_secret_chars

        return base64.b32encode(base32_secret_chars.encode("utf-8")).decode("utf-8")



# Generated by Django 4.2 on 2023-08-28 06:32

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="VerificationToken",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("created_date", models.DateTimeField(auto_now_add=True, verbose_name="Created date")),
                ("update_date", models.DateTimeField(auto_now=True, verbose_name="Update date")),
                ("token", models.CharField(max_length=6, verbose_name="Token")),
                (
                    "token_type",
                    models.CharField(
                        choices=[("totp", "Totp"), ("hotp", "Hotp")],
                        default="totp",
                        max_length=100,
                        verbose_name="Verify Type",
                    ),
                ),
                ("expire_date", models.DateTimeField(auto_now=True, null=True, verbose_name="Expire date")),
                ("is_active", models.BooleanField(default=True, verbose_name="Is active")),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL, verbose_name="User"
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="UserNotificationPreference",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("created_date", models.DateTimeField(auto_now_add=True, verbose_name="Created date")),
                ("update_date", models.DateTimeField(auto_now=True, verbose_name="Update date")),
                (
                    "notification_type",
                    models.CharField(
                        choices=[("email", "Email"), ("sms", "Sms"), ("authenticator", "Authenticator")],
                        default="email",
                        max_length=100,
                    ),
                ),
                (
                    "user",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="notification_preference",
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="User",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
    ]



import random
import string

from django.conf import settings
from django.contrib.auth.models import AbstractUser
from django.db import models
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

from .enums import NotificationType, TokenType


class AbstractBaseModel(models.Model):
    created_date = models.DateTimeField(auto_now_add=True, verbose_name=_("Created date"))
    update_date = models.DateTimeField(auto_now=True, verbose_name=_("Update date"))

    class Meta:
        abstract = True


class VerificationToken(AbstractBaseModel):
    user = models.ForeignKey(to=settings.AUTH_USER_MODEL, on_delete=models.CASCADE, verbose_name=_("User"))
    token = models.CharField(max_length=6, verbose_name=_("Token"))
    token_type = models.CharField(
        max_length=100, choices=TokenType.choices, default=TokenType.TOTP, verbose_name=_("Token type")
    )
    secret_chars = models.CharField(max_length=64, null=True, verbose_name=_("Extra secret characters"))
    is_active = models.BooleanField(default=True, verbose_name=_("Is active"))

    @staticmethod
    def generate_secret_chars(user: AbstractUser) -> str:
        alphanumeric_chars = string.ascii_letters + string.digits

        random_chars = "".join([random.choice(alphanumeric_chars) for i in range(10)])
        username_field_value = getattr(user, user.USERNAME_FIELD)
        timestamp = str(timezone.now().timestamp()).replace(".", "")

        secret_chars = username_field_value + random_chars + timestamp
        return secret_chars


class UserNotificationPreference(AbstractBaseModel):
    user = models.OneToOneField(
        to=settings.AUTH_USER_MODEL,
        on_delete=models.CASCADE,
        verbose_name=_("User"),
        related_name="notification_preference",
    )
    notification_type = models.CharField(
        max_length=32,
        choices=NotificationType.choices,
        default=NotificationType.EMAIL,
        verbose_name=_("Notification Type"),
    )



from django.apps import AppConfig


class RestFrameworkOTPConfig(AppConfig):
    name = "django_drf_otp"
    verbose_name = "Django REST framework OTP"
    default_auto_field = "django.db.models.BigAutoField"

    def ready(self) -> None:
        from .signals import user_is_created_or_save



# Generated by Django 4.2 on 2023-09-12 12:09

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_drf_otp', '0004_alter_usernotificationpreference_id_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='verificationtoken',
            name='secret_chars',
            field=models.CharField(max_length=64, null=True, verbose_name='Extra secret characters'),
        ),
        migrations.AlterField(
            model_name='usernotificationpreference',
            name='notification_type',
            field=models.CharField(
                choices=[('email', 'Email'), ('sms', 'Sms'), ('authenticator', 'Authenticator')],
                default='email',
                max_length=32,
                verbose_name='Notification Type',
            ),
        ),
        migrations.AlterField(
            model_name='verificationtoken',
            name='token_type',
            field=models.CharField(
                choices=[('totp', 'Totp'), ('hotp', 'Hotp')], default='totp', max_length=100, verbose_name='Token type'
            ),
        ),
    ]
